from typing import TypeVar, Type, Generic, Optional, Callable

from pyteamtv.infra.requester import Requester

T = TypeVar('T')


class List(Generic[T]):
    def __init__(self,
                 content_class: Type[T],
                 requester: Requester,
                 method: str,
                 url: str,
                 item_filter: Optional[Callable[[T], bool]] = None):
        self.requester = requester
        self.content_class = content_class
        self.url = url

        data = requester.request(method, url)
        items = [content_class(
            requester,
            item
        ) for item in data]
        if item_filter:
            self._items = [
                item for item in items if item_filter(item)
            ]
        else:
            self._items = items

    def __getitem__(self, index) -> T:
        assert isinstance(index, (int, slice))
        return self._items[index]

    def __iter__(self):
        for item in self._items:
            yield item

    # TODO: not used yet, might be usefull
    # when /api/sportingEvent/<uuid>/videos endpoint exists
    # def create(self, body) -> T:
    #     data = self.requester.request('POST', self.url, body)
    #     item = self.content_class(
    #         self.requester,
    #         data
    #     )
    #     self._items.append(item)
    #     return item






from pyteamtv.infra.requester import Requester
from .person import PersonResourceGroup
from .team import TeamResourceGroup
from .club import ClubResourceGroup
from .sharing_group import SharingGroupResourceGroup
from .reseller import ResellerResourceGroup
from .exchange import ExchangeResourceGroup
from .app_developer import AppDeveloperResourceGroup
from .user import UserResourceGroup


def factory(requester: Requester, attributes: dict):
    """

    :param requester:
    :param attributes:
    :rtype: Team
    """
    _CLASSES = {
        'team': TeamResourceGroup,
        'club': ClubResourceGroup,
        'SharingGroup': SharingGroupResourceGroup,
        'reseller': ResellerResourceGroup,
        'exchange': ExchangeResourceGroup,
        'app-developer': AppDeveloperResourceGroup,
        'user': UserResourceGroup,
        'person': PersonResourceGroup
    }

    target_resource_type, target_resource_id = attributes['targetResourceId'].split(":")

    if target_resource_type not in _CLASSES:
        raise Exception(f"Unknown resource group type {target_resource_type}")

    return _CLASSES[target_resource_type](
        requester.with_extra_headers({
            'X-Resource-Group-Id': attributes['resourceGroupId']
        }),
        attributes
    )



from ..teamtv_object import TeamTVObject
from .capabilities import _HasStorage


class _ResourceGroup(TeamTVObject, _HasStorage):
    def _use_attributes(self, attributes: dict):
        super()._use_attributes(attributes)

        self._name = attributes['targetResourceName']
        self._tenant_id = attributes['tenantId']
        self._resource_group_id = attributes['resourceGroupId']

    @property
    def resource_group_id(self):
        return self._resource_group_id

    @property
    def name(self):
        return self._name

    @property
    def tenant_id(self):
        return self._tenant_id

    def __repr__(self):
        return f"<ResourceGroup name='{self.name}'>"



from .teamtv_object import TeamTVObject


class Team(TeamTVObject):
    @property
    def name(self):
        return self._name

    @property
    def team_id(self):
        return self._team_id

    @property
    def sport_type(self):
        return self._sport_type

    @property
    def tags(self):
        return self._tags

    def _use_attributes(self, attributes: dict):
        self._team_id = attributes['teamId']
        self._name = attributes['name']
        self._sport_type = attributes['sportType']
        self._tags = attributes.get('tags', {})



from pyteamtv.models.sharing_group import SharingGroup
from . import _ResourceGroup

from .capabilities import _HasTeamsMixin, _HasSportingEventsMixin, _HasVideosMixin


class SharingGroupResourceGroup(_ResourceGroup, _HasTeamsMixin, _HasSportingEventsMixin, _HasVideosMixin):
    def get_sharing_group(self):
        return SharingGroup(
            self._requester,
            self._requester.request(
                "GET",
                f"/sharingGroups/{self.sharing_group_id}"
            )
        )

    @property
    def sharing_group_id(self):
        return self._sharing_group_id

    def _use_attributes(self, attributes: dict):
        self._sharing_group_id = attributes['targetResourceId'].split(":")[1]
        super()._use_attributes(attributes)






import time

import requests
import logging


logger = logging.getLogger(__name__)


class Requester(object):
    def __init__(self, base_url, jwt_token, headers:dict = None):
        self._base_url = base_url
        self.jwt_token = jwt_token
        self.headers = dict(**headers) if headers else {}

    def request(self, method, url, input_=None):
        import pyteamtv

        start = time.time()
        headers = self.headers
        headers['Authorization'] = f'Bearer {self.jwt_token}'
        headers['User-Agent'] = f'pyteamtv {pyteamtv.__version__}'
        logger.debug(f"Sending {method} request to {url} - {self.headers}")
        response = requests.request(
            method,
            self._base_url + url,
            headers=headers,
            json=input_
        )
        took = time.time() - start
        logger.debug(f"Request took: {took * 1000:.2f}ms")
        if response.status_code != 200:
            raise Exception(str(response))

        return response.json()

    def with_extra_headers(self, headers:dict):
        new_headers = dict(**self.headers)
        new_headers.update(headers)
        return self.__class__(
            self._base_url,
            self.jwt_token,
            headers
        )





from .teamtv_object import TeamTVObject
from .resource_group.factory import factory as resource_group_factory


class Membership(TeamTVObject):
    @property
    def role_names(self):
        return self._role_names

    @property
    def resource_group(self):
        return self._resource_group

    @property
    def member_attributes(self):
        return self._member_attributes

    def _use_attributes(self, attributes: dict):
        self._role_names = attributes['roleNames']
        self._member_attributes = attributes['memberAttributes']

        self._resource_group = resource_group_factory(
            self._requester,
            dict(
                tenantId=attributes['tenantId'],
                resourceGroupId=attributes['resourceGroupId'],
                targetResourceId=attributes['targetResourceId'],
                targetResourceName=attributes['targetResourceName']
            )
        )

    def __repr__(self):
        return f"<Membership resourceGroup={self.resource_group} role={self.role_names[0]}>"



from dataclasses import dataclass

from .teamtv_object import TeamTVObject


@dataclass
class Credentials(object):
    access_key_id: str
    secret_access_key: str
    session_token: str


@dataclass
class BucketInfo(object):
    bucket: str
    base_key: str


class AppStorageTokens(TeamTVObject):
    @property
    def credentials(self):
        return self._credentials

    def _use_attributes(self, attributes: dict):
        self._credentials = Credentials(
            access_key_id=attributes['credentials']['accessKeyId'],
            secret_access_key=attributes['credentials']['secretAccessKey'],
            session_token=attributes['credentials']['sessionToken']
        )
        self._bucket_info = BucketInfo(
            bucket=attributes['storage']['bucket'],
            base_key=attributes['storage']['baseKey']
        )



from typing import Optional

import jwt

from pyteamtv.infra.requester import Requester
from pyteamtv.models.list import List
from pyteamtv.models.membership import Membership
from pyteamtv.models.membership_list import MembershipList
from pyteamtv.models.sharing_group import SharingGroup

from .endpoint import API_ENDPOINT
from .token import TOKEN
from ..exceptions import TeamNotFound, InputError
from ..models.resource_group.team import TeamResourceGroup


class TeamTVUser(object):
    def __init__(self, jwt_token):
        self.jwt_token = jwt_token

        token = jwt.decode(
            jwt_token,
            TOKEN,
            algorithms='RS256',
            verify=False,
            options={
                'verify_signature': False
            }
        )

        self._requester = Requester(
            f"{API_ENDPOINT}/api",
            jwt_token
        )
        self.__token = token

    def get_membership_list(self):
        """
        :return: MembershipList
        """
        return MembershipList(
            Membership,
            self._requester,
            "GET",
            "/users/me/memberships"
        )

    def get_public_sharing_groups(self):
        return List(
            SharingGroup,
            self._requester,
            "GET",
            "/sharingGroups"
        )

    def get_team(self, name: Optional[str] = None, resource_group_id: Optional[str] = None) -> TeamResourceGroup:
        membership_list = self.get_membership_list()
        if resource_group_id:
            membership = membership_list.get_membership_by_resource_group_id(
                resource_group_id
            )
        elif name:
            membership = membership_list.get_membership_by_name(
                name=name
            )
        else:
            raise InputError(f"Either `name` or `resource_group_id` must be specified.")

        if not membership:
            raise TeamNotFound(f"No team named '{name}' found.")
        return membership.resource_group





import logging
import os
from datetime import datetime
from typing import Optional

from .list import List
from .observation import Observation
from .observation_log import ObservationLog
from .teamtv_object import TeamTVObject
from .video import Video

from tusclient.uploader import Uploader

class SportingEvent(TeamTVObject):
    @property
    def scheduled_at(self) -> datetime:
        return self._scheduled_at

    @property
    def sporting_event_id(self) -> str:
        return self._sporting_event_id

    @property
    def name(self):
        return self._name

    @property
    def tags(self):
        return self._tags

    @property
    def main_video_id(self) -> Optional[str]:
        if self._video_ids:
            return self._video_ids[0]
        return None

    @property
    def is_local(self):
        return self._is_local

    def get_observation_log(self) -> ObservationLog:
        video_id = self.main_video_id
        if video_id:
            clock_id = self._clocks[video_id]['clockId']
        else:
            clock_id = 'U1'

        return ObservationLog(
            Observation,
            self._requester,
            "GET",
            f"/sportingEvents/{self.sporting_event_id}/observations/{clock_id}",
            clock_id,
            self
        )

    def get_videos(self) -> List[Video]:
        def _filter(video: Video) -> bool:
            return video.video_id in self._video_ids

        return List(
            Video,
            self._requester,
            "GET",
            "/videos",
            item_filter=_filter
        )

    def upload_video(self, *file_paths: str, chunks_per_request: int = 1) -> Video:
        chunks_per_request = int(chunks_per_request)
        if chunks_per_request < 1:
            chunks_per_request = 1

        files = []
        for filename in file_paths:
            files.append({
                'name': os.path.basename(filename),
                'size': os.path.getsize(filename)
            })

        video_id = self._requester.request(
            "POST",
            f"/sportingEvents/{self.sporting_event_id}/initVideoUpload",
            {
                "files": files
            }
        )

        video = Video(
            self._requester,
            self._requester.request(
                "GET",
                f"/videos/{video_id}"
            )
        )
        for i, part in enumerate(video.parts):
            uploader = Uploader(
                file_paths[i],
                url=part['tusUploadUrl'],
                chunk_size=chunks_per_request * 25 * 1024 * 1024,
                log_func=lambda msg: logging.info(f"Uploading: {msg}"),

                retries=5,
                retry_delay=5
            )

            uploader.upload()

        return Video(
            self._requester,
            self._requester.request(
                "GET",
                f"/videos/{video_id}"
            )
        )

    def _use_attributes(self, attributes: dict):
        self._name = attributes['name']
        self._sporting_event_id = attributes['sportingEventId']
        self._tags = attributes.get('tags', {})
        self._video_ids = attributes.get('videoIds', [])
        self._clocks = attributes['clocks']
        self._scheduled_at = datetime.fromisoformat(attributes['scheduledAt'].replace("Z", "+00:00"))
        self._is_local = attributes['_metadata']['source']['type'] == 'ResourceGroup'

    def __str__(self):
        return f"{self._scheduled_at.strftime('%d/%m/%y')} {self._name}"



from .teamtv_object import TeamTVObject


class Observation(TeamTVObject):
    @property
    def observation_id(self) -> str:
        return self._observation_id

    @property
    def start_time(self) -> float:
        return self._start_time

    @property
    def trigger_time(self) -> float:
        return self._trigger_time

    @property
    def end_time(self) -> float:
        return self._end_time

    @property
    def code(self) -> str:
        return self._code

    @property
    def clock_id(self) -> str:
        return self._clock_id

    @property
    def description(self) -> str:
        return self._description

    @property
    def attributes(self) -> dict:
        return self._attributes

    def _use_attributes(self, attributes: dict):
        self._observation_id = attributes['observationId']
        self._start_time = attributes['startTime']
        self._trigger_time = attributes['triggerTime']
        self._end_time = attributes['endTime']
        self._code = attributes['code']
        self._attributes = attributes['attributes'] or dict()
        self._description = attributes['description']
        self._clock_id = attributes['clockId']



from typing import Type, Set, TYPE_CHECKING
import math

if TYPE_CHECKING:
    from pyteamtv.models.sporting_event import SportingEvent

from pyteamtv.infra.requester import Requester
from pyteamtv.models.list import List
from pyteamtv.models.observation import Observation


def pol2cart(angle: float, distance: float):
    if angle > 90:
        angle -= 360
    angle += 90

    phi = (angle / 360) * 2 * math.pi
    rho = distance

    x = rho * math.cos(phi)
    y = rho * math.sin(phi)
    return x, y


class ObservationLog(List[Observation]):
    def __init__(self,
                 content_class: Type[Observation],
                 requester: Requester,
                 method: str,
                 url: str,
                 clock_id: str,
                 sporting_event: 'SportingEvent'
                 ):
        super().__init__(content_class, requester, method, url)
        self._clock_id = clock_id
        self._sporting_event = sporting_event

    def get_mapping_stats(self):
        stats = dict(
            success=0,
            failed=0
        )
        for observation in self:
            if observation.clock_id == self._clock_id:
                stats['success'] += 1
            else:
                stats['failed'] += 1
        return stats

    def to_pandas(self):
        try:
            import pandas as pd
        except ImportError:
            raise Exception("You don't have pandas installed. Please install first")

        observations = []
        skip_attributes = {"teamId", "personId", "team", "person", "inPerson", "inPersonId", "outPerson", "outPersonId"}
        team = dict()
        for observation in self:
            if observation.code == 'START-POSSESSION':
                team = dict(
                    team_id=observation.attributes['teamId'],
                    team_name=observation.attributes['team']['name'],
                    team_ground=observation.attributes['team']['ground'],
                    position=observation.attributes['position']
                )
            else:
                if 'person' in observation.attributes:
                    person = dict(
                        person_id=observation.attributes['personId'],
                        first_name=observation.attributes['person']['firstName'],
                        last_name=observation.attributes['person']['lastName'],
                        number=observation.attributes['person']['number']
                    )
                    person['full_name'] = person['first_name'] + ' ' + person['last_name']
                else:
                    person = dict()

                attributes = dict()
                attributes.update(team)
                attributes.update(person)
                for k, v in observation.attributes.items():
                    if k not in skip_attributes:
                        attributes[k] = v

                if 'angle' in attributes and 'distance' in attributes:
                    attributes['x'], attributes['y'] = pol2cart(
                        attributes['angle'], attributes['distance']
                    )

                observation_dict = dict(
                    sporting_event_id=self._sporting_event.sporting_event_id,
                    sporting_event_name=self._sporting_event.name,
                    sporting_event_scheduled_at=self._sporting_event.scheduled_at,
                    observation_id=observation.observation_id,
                    clock_id=observation.clock_id,
                    start_time=observation.start_time,
                    end_time=observation.end_time,
                    code=observation.code,
                    description=observation.description,
                    **attributes
                )

                observations.append(observation_dict)

        return pd.DataFrame.from_records(observations)



from datetime import datetime

from pyteamtv.infra.requester import Requester
from ..list import List

from ..sporting_event import SportingEvent
from ..team import Team
from ..video import Video
from ..exercise import Exercise
from ..app_storage import AppStorageTokens


def iso8601(datetime_: datetime):
    assert datetime_.tzname() == "UTC", datetime_

    return datetime_.strftime("%Y-%m-%dT%H:%M:%S.%fZ")


class BaseMixin(object):
    _requester: Requester


class _HasStorage(BaseMixin):
    def get_app_storage_tokens(self):
        return AppStorageTokens(
            self._requester,
            self._requester.request(
                "GET",
                "/apps/current/storageTokens"
            )
        )


class _HasTeamsMixin(BaseMixin):
    def get_teams(self):
        return List(
            Team,
            self._requester,
            "GET",
            "/teams"
        )

    def create_team(self, name: str, sport_type: str, tags: dict = None):
        data = self._requester.request(
            "POST",
            "/teams",
            dict(
                name=name,
                sportType=sport_type,
                tags=tags if tags else {}
            )
        )
        return Team(
            self._requester,
            data
        )


class _HasExercisesMixin(BaseMixin):
    def get_exercises(self):
        return List(
            Exercise,
            self._requester,
            "GET",
            "/exercises"
        )

    def create_exercise(self, title: str, exercise_template_id: str = "exercise-template-sv", tags: dict = None):
        data = self._requester.request(
            "POST",
            "/exercises",
            dict(
                title=title,
                exerciseTemplateId=exercise_template_id,
                tags=tags if tags else {}
            )
        )
        return Exercise(
            self._requester,
            data
        )


class _HasSharingGroupResourceGroupsMixin(BaseMixin):
    def get_sharing_group_resource_groups(self):
        from .factory import factory

        return List(
            factory,
            self._requester,
            "GET",
            "/resourceGroups"
        )

    def create_sharing_group_resource_group(self, name: str, description: str, sport_type: str, visibility_type: str = 'public', attributes:dict = None):
        from .factory import factory

        data = self._requester.request(
            "POST",
            "/resourceGroups",
            dict(
                name=name,
                sportType=sport_type,
                description=description,
                visibility_type=visibility_type,
                attributes=attributes if attributes else {}
            )
        )
        return factory(
            self._requester,
            data
        )


class _HasSportingEventsMixin(BaseMixin):
    def get_sporting_events(self):
        return List[SportingEvent](
            SportingEvent,
            self._requester,
            "GET",
            "/sportingEvents"
        )

    def get_sporting_event(self, sporting_event_id: str):
        data = self._requester.request(
            "GET",
            f"/sportingEvents/{sporting_event_id}"
        )
        return SportingEvent(
            self._requester,
            data
        )

    def create_sporting_event(self,
                              home_team: Team,
                              away_team: Team,
                              scheduled_at: datetime,
                              tags: dict = None,
                              type: str = "match"):
        data = self._requester.request(
            "POST",
            "/sportingEvents",
            dict(
                type=type,
                name=f"{home_team.name} - {away_team.name}",
                homeTeamId=home_team.team_id,
                awayTeamId=away_team.team_id,
                scheduledAt=iso8601(scheduled_at),
                tags=tags if tags else {}
            )
        )
        return SportingEvent(
            self._requester,
            data
        )


class _HasVideosMixin(BaseMixin):
    def get_videos(self):
        return List(
            Video,
            self._requester,
            "GET",
            "/videos"
        )

    def get_video(self, video_id):
        data = self._requester.request(
            "GET",
            f"/videos/{video_id}"
        )

        return Video(
            self._requester,
            data
        )



from . import _ResourceGroup

from .capabilities import _HasTeamsMixin, _HasSportingEventsMixin, _HasVideosMixin


class ResellerResourceGroup(_ResourceGroup, _HasTeamsMixin, _HasSportingEventsMixin, _HasVideosMixin):
    pass



import jwt

from pyteamtv.infra.requester import Requester
from pyteamtv.models.resource_group.factory import factory as resource_group_factory
from .endpoint import API_ENDPOINT

from .token import TOKEN


class TeamTVApp(object):
    def __init__(self, jwt_token, app_id: str):
        self.jwt_token = jwt_token

        token = jwt.decode(
           jwt_token,
           TOKEN,
           algorithms='RS256',
           verify=True,
           audience=f"app:{app_id}"
        )

        self._requester = Requester(
            f"{API_ENDPOINT}/api",
            jwt_token
        )
        self.__token = token

    def get_resource_group(self):
        """
        :rtype: :class:`pyteamtv.resource_group.team.Team`
        """
        data = self._requester.request(
           "GET",
           "/resourceGroups/current"
        )

        return resource_group_factory(
           self._requester,
           data
        )



from typing import Optional

from .list import List
from .membership import Membership


class MembershipList(List):
    def get_memberships_of_types(self, class_):
        return list(
            filter(
                lambda membership: isinstance(membership.resource_group, class_),
                self._items
            )
        )

    def get_membership_by_resource_group_id(self, resource_group_id: str) -> Membership:
        for membership in self._items:
            if membership.resource_group.resource_group_id == resource_group_id:
                return membership
        return None

    def get_membership_by_name(self, name: str) -> Membership:
        for membership in self._items:
            if membership.resource_group.name == name:
                return membership
        return None

